// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "UILabel.hpp"
#include "game/DawnGame.hpp"

using namespace Dawn;

UILabel::UILabel(SceneItem *item) :
  UIComponentRenderable(item)
{

}

void UILabel::onStart() {
  this->shaderBuffer.init();
}

std::vector<struct ShaderPassItem> UILabel::getUIRenderPasses() {
  // if(this->texts.size() == 0) return {};

  auto canvas = this->getCanvas();
  auto shader = getGame()->renderManager.fontShader;

  struct ShaderPassItem item;
  item.shader = shader;
  item.mesh = &this->mesh;
  item.matrixValues[shader->paramModel] = transform->getWorldTransform();
  item.parameterBuffers[shader->bufferUiCanvas] = &canvas->shaderBuffer;
  item.parameterBuffers[shader->bufferFont] = &this->shaderBuffer;
  item.renderFlags = RENDER_MANAGER_RENDER_FLAG_BLEND;
  item.start = quadStart * QUAD_VERTICE_COUNT;
  item.count = quadCount == -1 ? -1 : quadCount * QUAD_VERTICE_COUNT;

  // Map texture slots
  auto it = textureMap.begin();
  while(it != textureMap.end()) {
    shaderparameter_t param;
    switch(it->second) {
      case 0:
        param = shader->paramTexture0;
        break;
      
      case 1:
        param = shader->paramTexture1;
        break;

      case 2:
        param = shader->paramTexture2;
        break;

      case 3:
        param = shader->paramTexture3;
        break;

      default:
        assertUnreachable();
    }

    item.textureSlots[it->second] = &it->first->texture;
    item.textureValues[param] = it->second;
    ++it;
  }

  return { item };
}

float_t UILabel::getContentWidth() {
  return 0;
}

float_t UILabel::getContentHeight() {
  return 0;
}

void UILabel::rebufferQuads(std::vector<struct UILabelText> newTexts) {
  auto oldTexts = this->texts;

  textureMap.clear();
  glm::vec2 position(0, 0);
  struct FontShaderBufferData fontData;
  int32_t quadIndex = 0;
  int32_t partIndex = 0;
  quadCountTotal = 0;
  int32_t nextTexture = 0;

  // Determine how many quads there are, and the texture indexes.
  auto itText = newTexts.begin();
  while(itText != newTexts.end()) {
    quadCountTotal += itText->text.length();

    // Determine font and lock it.
    assertNotNull(itText->style.font);
    itText->lockId = itText->style.font->lock(NewTrueTypeFaceTextureStyle{
      itText->style.size,
      itText->style.style
    });
    assertTrue(itText->lockId != -1);
    itText->texture = itText->style.font->getTexture(itText->lockId);

    // Check for existing texture, if not, map it.
    if(textureMap.find(itText->texture) == textureMap.end()) {
      assertTrue(nextTexture < FONT_SHADER_TEXTURE_MAX);
      textureMap[itText->texture] = nextTexture++;
    }

    ++itText;
  }

  // Cleanup old texst, we do this second so we don't unlock, cleanup, and then
  // lock the same font, causing it to have to re-load.
  itText = oldTexts.begin();
  while(itText != oldTexts.end()) {
    assertTrue(itText->lockId != -1);
    assertNotNull(itText->style.font);
    itText->style.font->unlock(itText->lockId);
    ++itText;
  }

  // Update texts.
  this->texts = newTexts;
  
  // Create mesh
  this->mesh.createBuffers(
    QUAD_VERTICE_COUNT * quadCountTotal,
    QUAD_INDICE_COUNT * quadCountTotal
  );

  // Buffer the text quads
  itText = newTexts.begin();
  while(itText != newTexts.end()) {
    position.y += itText->style.size;
    quadIndex += this->bufferQuads(
      *itText,
      fontData,
      textureMap,
      position,
      quadIndex,
      partIndex
    );
    // position.y -= itText->style.size;
    ++partIndex;
    ++itText;
  }

  shaderBuffer.buffer(&fontData);
}

int32_t UILabel::bufferQuads(
  struct UILabelText text,
  struct FontShaderBufferData &bufferData,
  std::map<NewTrueTypeFaceTexture*, int32_t> &textureMap,
  glm::vec2 &position,
  int32_t quadStart,
  int32_t partIndex
) {
  // Get string length
  int32_t len = text.text.length();
  
  glm::vec2 wh = glm::vec2(
    text.texture->texture.getWidth(),
    text.texture->texture.getHeight()
  );

  // For each char
  int32_t lastSpaceCharacter = -1;
  for(int32_t i = 0; i < len; i++) {
    char ch = text.text[i];

    if(ch == '\n') {
      position.x = 0;
      position.y += text.style.size;
      ch = ' ';
      lastSpaceCharacter = i;
    } else if(ch == ' ') {
      lastSpaceCharacter = i;
    }
    
    // Invalid/Unsupported chars
    assertTrue(ch >= NEW_TRUETYPE_CHAR_BEGIN && ch < NEW_TRUETYPE_CHAR_END);
    assertTrue(ch != '\r');
    assertTrue(ch != '\t');

    int32_t j = quadStart + i;
    FT_ULong c = ch;
    auto charInfo = text.texture->getCharacterData(c);

    // Word wrapping
    if(
      lastSpaceCharacter != -1 &&
      this->width > 0 &&
      (position.x+charInfo.advanceX) > this->width
    ) {
      text.text[lastSpaceCharacter] = '\n';
      i = lastSpaceCharacter - 1;
      lastSpaceCharacter = -1;
      continue;
    }

    // Determine texture coordinates.
    glm::vec2 uv0 = glm::vec2(0.0f, charInfo.textureY) / wh;
    glm::vec2 uv1 = uv0 + (charInfo.bitmapSize / wh);

    // Buffer the quad.
    assertTrue(j < FONT_SHADER_QUADS_MAX);
    QuadMesh::bufferQuadMeshWithZ(&this->mesh,
      position + charInfo.bitmapPosition, uv0,
      position + charInfo.bitmapPosition + charInfo.bitmapSize, uv1,
      0.0f,
      j * QUAD_VERTICE_COUNT, j * QUAD_INDICE_COUNT
    );

    // Move the current position along.
    position.x += charInfo.advanceX;
    position.y += charInfo.advanceY;

    // Set the part index to the quad mappings
    bufferData.fontQuadMappings[j] = partIndex;
  }

  // Map texture level values
  auto textureId = textureMap.find(text.texture);
  assertTrue(textureId != textureMap.end());
  bufferData.textures[partIndex] = textureId->second;
  bufferData.colors[partIndex] = text.style.color;

  return len;
}